{% if fluff -%}/**********************************************************************************************
 * Filename:       app_snippets.c
 *
 * Description:    This file contains snippets needed to utilize the generated services.
 *
 *                 Generated by:
 *                 BDS version: {{pluginInfo.AppVersion}}
 *                 Plugin:      {{pluginInfo.Name}} {{ pluginInfo.Version}}
 *                 Time:        {{pluginInfo.timestamp}}
 *
 * Copyright (c) 2015, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *************************************************************************************************/
{% endif -%}
//...
#include <ti/sysbios/knl/Semaphore.h>
#include <ti/sysbios/knl/Queue.h>

//...
{% for service in services -%}
#include "{{service.name}}.h"
{% endfor -%}

//...

{% if anyWriteable or anyCCCDs -%}
// Types of messages that can be sent to the application task
typedef enum
{
  APP_MSG_SERVICE_WRITE = 0,
  APP_MSG_SERVICE_CFG,
  APP_MSG_SEND_NOTIFICATION,
  APP_MSG_GAP_STATE_CHANGE
} app_msg_types_t;

// Struct for messages sent to the application task
typedef struct
{
  Queue_Elem       _elem;
  app_msg_types_t  type;
  uint8_t          pdu[];
} app_msg_t;

// Struct for messages about characteristic data
typedef struct
{
  uint16_t svcUUID; // UUID of the service
  uint16_t dataLen; // 
  uint8_t  paramID; // Index of the characteristic
  uint8_t  data[];  // Flexible array member, extended to malloc - sizeof(.)
} char_data_t;

// Semaphore globally used to post events to the application thread
static ICall_Semaphore sem; // Already present in sample applications.

// Queue object used for application messages.
static Queue_Struct applicationMsgQ;
static Queue_Handle hApplicationMsgQ;

{% endif -%}

{% if anyCCCDs -%}
// Clock structs for periodic notification example
{% for service in services -%}
{% for char in service.charsHaveCCCDs -%}
static Clock_Struct {{ service.nameAbbr }}_{{ char.nameCap }}_clock;
{% endfor -%}
{% endfor -%}

{% endif -%}
//...

// Declaration of BDS generated service callback handlers
{% for service in services -%}
{% if service.writeable -%}
static void user_{{service.nameCap}}_ValueChangeCB( uint16_t connHandle, uint8_t paramID, uint8_t *pValue, uint16_t len ); // Callback from the service on writes.
static void user_{{service.nameCap}}_ValueChangeDispatchHandler(char_data_t *pCharData); // Local handler called from the Task context of this task.
{% else -%}
// {{service.name}} has no writeable characteristics
{% endif -%}
{% if service.numCCCDs -%}
static void user_{{service.nameCap}}_CfgChangeCB( uint16_t connHandle, uint8_t paramID, uint8_t *pValue, uint16_t len ); // Callback from the service on noti/ind config change.
static void user_{{service.nameCap}}_CfgChangeDispatchHandler(char_data_t *pCharData);
{% endif -%}
{% endfor -%}
{% if anyCCCDs -%}

// Callback handlers for the clock objects
{% for service in services -%}
{%   if service.numCCCDs -%}
static void user_{{service.nameCap}}_clockSwiHandler(UArg paramID);
{%   endif -%}
{% endfor -%}

// Task handler for sending notifications.
static void user_updateCharVal(char_data_t *pCharData);
{% endif -%}

static void user_enqueueRawAppMsg( app_msg_types_t appMsgType, uint8_t *pData, uint16_t len );
{% if anyWriteable or anyCCCDs -%}
static void user_enqueueCharDataMsg( app_msg_types_t appMsgType, uint16_t serviceUUID, 
                                       uint16_t connHandle, uint8_t paramID, 
                                       uint8_t *pValue, uint16_t len );

{% endif -%}

// GAP Role Callbacks
static gapRolesCBs_t user_gapRoleCBs =
{
  user_gapStateChangeCB     // Profile State Change Callbacks
};

// GAP Bond Manager Callbacks
static gapBondCBs_t user_bondMgrCBs =
{
  NULL, // Passcode callback (not used by application)
  NULL  // Pairing / Bonding state Callback (not used by application)
};


// Service callback structure for registering our handlers with the service
{% for service in services -%}
// {{ service.nameNice }} callback handler. The type {{service.name}}CBs_t is defined in {{service.name}}.h
static {{service.nameCap}}CBs_t user_{{service.name}}CBs =
{
{% if service.writeable -%}
  .pfnChangeCb    = user_{{service.nameCap}}_ValueChangeCB, // Characteristic value change callback handler
{% else -%}
  .pfnChangeCb    = NULL, // No writable chars in {{ service.nameNice }}, so no change handler.
{% endif -%}
{% if service.numCCCDs -%}
  .pfnCfgChangeCb = user_{{service.nameCap}}_CfgChangeCB, // Noti/ind configuration callback handler
{% else -%}
  .pfnCfgChangeCb = NULL, // No notification-/indication enabled chars in {{ service.nameNice }}
{% endif -%}
};

{% endfor -%}

//...



// taskFxn_init(..)
// {
{% if anyCCCDs -%}
  Clock_Params clockParams;
  Clock_Params_init(&clockParams);
  clockParams.period = 5000 * (1000 / Clock_tickPeriod); // 5000ms period

  // Clock struct initialization for periodic notification example
  // Clock callbacks only have one parameter, so make one callback handler per service.
{% for service in services -%}
{% for char in service.charsHaveCCCDs -%}
  clockParams.arg = {{ service.nameAbbr | upcase }}_{{ char.nameCap | upcase }}_ID;
  Clock_construct(&{{ service.nameAbbr }}_{{ char.nameCap }}_clock,
                  user_{{service.nameCap}}_clockSwiHandler,
                  0,
                  &clockParams);
{% endfor -%}
{% endfor -%}

{% endif -%}
{% if anyWriteable -%}
  // Initialize queue for application messages.
  // Note: Used to transfer control to application thread
  Queue_construct(&applicationMsgQ, NULL);
  hApplicationMsgQ = Queue_handle(&applicationMsgQ);
{% endif -%}



  // Add services to GATT server
{% for service in services -%}
  {{ service.nameCap }}_AddService( selfEntity );
{% if service.writeable or service.numCCCDs -%}
  {{ service.nameCap }}_RegisterAppCBs(&user_{{service.name}}CBs);
{% endif -%}
{% endfor -%}

  // Placeholder variable for characteristic intialization
  uint8_t someVal[20] = {0};

{% for service in services -%}
  // Initalization of characteristics in {{service.name}} that are readable.
{% for char in service.charsReadable -%}
  {{ service.nameCap }}_SetParameter({{ service.nameAbbr | upcase }}_{{ char.name | upcase }}_ID, {{ service.nameAbbr | upcase }}_{{ char.name | upcase }}_LEN, &someVal);
{% endfor -%}

{% endfor -%}
//}


// taskFxn(..)
// {
//   for (;;) {
//     ... ICall_wait(...)   // Wait for semaphore post
// 
{% if anyWriteable or anyCCCDs -%}
      // Process messages sent from another task or another context.
      while (!Queue_empty(hApplicationMsgQ))
      {
        app_msg_t   *pMsg = Queue_dequeue(hApplicationMsgQ);
        char_data_t *pCharData = (char_data_t *)pMsg->pdu;
        
        switch (pMsg->type)
        {
{%   if anyWriteable -%}
          case APP_MSG_SERVICE_WRITE: /* Message about received value write */
            /* Call different handler per service */
            switch(pCharData->svcUUID) {
{%     for service in services -%}
{%       if service.writeable -%}
              case {{ service.name | upcase }}_SERV_UUID:
                user_{{service.nameCap}}_ValueChangeDispatchHandler(pCharData);
                break;
{%       endif -%}
{%     endfor -%}
            }
          break;
{%   endif -%}
{%   if anyCCCDs -%}

          case APP_MSG_SERVICE_CFG: /* Message about received CCCD write */
            /* Call different handler per service */
            switch(pCharData->svcUUID) {
{%     for service in services -%}
{%       if service.numCCCDs -%}
              case {{ service.name | upcase }}_SERV_UUID:
                user_{{service.nameCap}}_CfgChangeDispatchHandler(pCharData);
                break;
{%       endif -%}
{%     endfor -%}
            }
          break;

          case APP_MSG_SEND_NOTIFICATION: /* Message to self from ClockSwi */
            user_updateCharVal(pCharData);
            break;
{%   endif -%}
        }

        // Free the received message.
        ICall_free(pMsg);
      }
{% else -%}
//   No writeable characteristics, so queue handling is not included.
{% endif -%}

//   }
// }


/******************************************************************************
 *
 *  Handlers of system events deferred to the user Task context.
 *
 *****************************************************************************/

{% for service in services -%}
{%   if service.writeable -%}

/**
 * Handle received characteristic write in task context.
 */
void user_{{service.nameCap}}_ValueChangeDispatchHandler(char_data_t *pCharData)
{
  switch (pCharData->paramID)
  {
{%     for char in service.charsNeedWrite -%}
    case {{ service.nameAbbr | upcase }}_{{ char.name | upcase }}_ID:
      // Do something useful with pCharData->data here
      // -------------------------
      break;
{%     endfor -%}
  }

}
{%   endif -%}
{%   if service.numCCCDs -%}

/**
 * Handle received characteristic configuration change in task context..
 */
void user_{{service.nameCap}}_CfgChangeDispatchHandler(char_data_t *pCharData)
{
  // Cast received data to uint16, as that's the format for CCCD writes.
  uint16_t configValue = *(uint16_t *)pCharData->data;

  switch (pCharData->paramID)
  {
{%     for char in service.charsHaveCCCDs -%}
    case {{ service.nameAbbr | upcase }}_{{ char.name | upcase }}_ID:
      // Do something useful with configValue here
      // -------------------------
      if (configValue) // 0x0001 and 0x0002 both indicate turned on.
        Clock_start((Clock_Handle)&{{ service.nameAbbr }}_{{ char.nameCap }}_clock);
      else
        Clock_stop((Clock_Handle)&{{ service.nameAbbr }}_{{ char.nameCap }}_clock);
      break;
{%     endfor -%}
  }
}
{%   endif -%}
{% endfor -%}
{% if anyCCCDs or anyReadable -%}

/**
 * Convenience function for updating characteristic data via char_data_t message
 * MUST run in Task context.
 */
static void user_updateCharVal(char_data_t *pCharData)
{
  switch(pCharData->svcUUID) {
{% for service in services -%}
{%   if service.numCCCDs or service.numReadable -%}
    case {{ service.name | upcase }}_SERV_UUID:
      {{ service.nameCap }}_SetParameter( pCharData->paramID, pCharData->dataLen, pCharData->data );
    break;

{%   endif -%}
{% endfor -%}
  }
}
{% endif -%}


/******************************************************************************
 *
 *  Handlers for callbacks occurring in other run-time contexts
 *  than the user Task context.
 *
 *****************************************************************************/

/*
 *  Callbacks from the Stack Task context
 *****************************************************************************/


{% for service in services -%}
{%   if service.writeable -%}

/**
 * Callback handler for characteristic value changes in {{service.nameNice}}, registered with the service.
 */
static void user_{{service.nameCap}}_ValueChangeCB(uint16_t connHandle, uint8_t paramID, uint8_t *pValue, uint16_t len )
{
  // See {{service.name}}.h to compare paramID with characteristic value attribute.
  user_enqueueCharDataMsg(APP_MSG_SERVICE_WRITE, {{ service.name | upcase }}_SERV_UUID, connHandle, paramID, pValue, len);
}
{%   endif-%}
{%   if service.numCCCDs -%}

/**
 * Callback handler for characteristic configuration changes in {{service.nameNice}}, registered with the service.
 */
static void user_{{service.nameCap}}_CfgChangeCB( uint16_t connHandle, uint8_t paramID, uint8_t *pValue, uint16_t len )
{
  user_enqueueCharDataMsg(APP_MSG_SERVICE_CFG, {{ service.name | upcase }}_SERV_UUID, connHandle, paramID, pValue, len);
}
{%   endif -%}
{% endfor -%}


/*
 *  Callbacks from Swi-context
 *****************************************************************************/
{% for service in services -%}
{%   if service.numCCCDs -%}
static void user_{{service.nameCap}}_clockSwiHandler(UArg paramID)
{
  uint8_t notiData[20];
  uint16_t notiLen = sizeof notiData;
{% if 0 -%}

This is an attempt to use Liquid syntax to determine whether a variable is needed.
It is not straightforward.

{% else -%}
{% assign anyWrite = 0 -%}
{%   for char in service.charsHaveCCCDs -%}
{%     if char.isWritable -%}
{%       assign anyWrite = 1 -%}
{%     endif -%}
{%   endfor -%}
{% if anyWrite == 0 -%}
  static uint8_t someCounter = 0;
{% endif -%}
{% endif -%}

  // Get loopback data if char is writable, otherwise fill with junk
  switch(paramID) {
{%     for char in service.charsHaveCCCDs -%}
    case {{ service.nameAbbr | upcase }}_{{ char.name | upcase }}_ID:
{%       if char.isWritable -%}
      // Characteristic is writable, so echo back contents.
      {{ service.nameCap }}_GetParameter( paramID, &notiLen, notiData );
{%       else -%}
      // Characteristic is not writable, so send something, max 20 bytes (max default ATT MTU size minus header)
      notiLen = MIN(notiLen, {{ service.nameAbbr | upcase }}_{{ char.name | upcase }}_LEN);
      memset(notiData, someCounter++, notiLen);
{%       endif -%}
    break;
{%     endfor -%}
  }

  // Send message to application that paramID (argument bound to the Clock_Struct) of
  // {{service.nameNice}} should emit a notification or indication.
  user_enqueueCharDataMsg(APP_MSG_SEND_NOTIFICATION, {{ service.name | upcase }}_SERV_UUID, 0xFFFF, paramID, notiData, notiLen);
}
{%   endif -%}
{% endfor -%}




/******************************************************************************
 *
 *  Utility functions
 *
 *****************************************************************************/

{% if anyWriteable or anyCCCDs -%}
/**
 * Generic message constructor for characteristic data.
 * Sends a message to the application for handling in Task context.
 */
static void user_enqueueCharDataMsg( app_msg_types_t appMsgType, uint16_t serviceUUID, 
                                       uint16_t connHandle, uint8_t paramID, 
                                       uint8_t *pValue, uint16_t len )
{
  // Called in Stack Task context, so can't do processing here.
  // Send message to application message queue about received data.
  uint16_t readLen = len; // How much data was written to the attribute
  
  // Allocate memory for the message.
  // Note: The pCharData message doesn't have to contain the data itself, as that's stored in
  //       a variable in the service. However, to prevent data loss if a new value is received
  //       before GetParameter is called, we call GetParameter now.
  app_msg_t *pMsg = ICall_malloc( sizeof(app_msg_t) + sizeof(char_data_t) + readLen );

  if (pMsg != NULL)
  {
    pMsg->type = appMsgType;
    
    char_data_t *pCharData = pMsg->pdu;
    pCharData->svcUUID = serviceUUID; // Use 16-bit part of UUID.
    pCharData->paramID = paramID;
    // Copy data from service now, because it could change before GetParameter is called later.
    memcpy(pCharData->data, pValue, readLen);
    // Update pCharData with how much data we received.
    pCharData->dataLen = readLen;
    // Enqueue the message using pointer to queue node element.
    Queue_enqueue(hApplicationMsgQ, &pMsg->_elem);
    // Let application know there's a message.
    Semaphore_post(sem);
  }
}

{% endif -%}
/**
 * Generic message constructor for raw data.
 * Sends a message to the application for handling in Task context.
 */
static void user_enqueueRawAppMsg( app_msg_types_t appMsgType, uint8_t *pData, uint16_t len );
{
  // Allocate memory for the message.
  // Note: The pCharData message doesn't have to contain the data itself, as that's stored in
  //       a variable in the service. However, to prevent data loss if a new value is received
  //       before GetParameter is called, we call GetParameter now.
  app_msg_t *pMsg = ICall_malloc( sizeof(app_msg_t) + len );

  if (pMsg != NULL)
  {
    pMsg->type = appMsgType;
    
    // Copy data from service now, because it could change before GetParameter is called later.
    memcpy(pMsg->pdu, pData, len);
    // Update pCharData with how much data we received.
    pCharData->dataLen = len;
    // Enqueue the message using pointer to queue node element.
    Queue_enqueue(hApplicationMsgQ, &pMsg->_elem);
    // Let application know there's a message.
    Semaphore_post(sem);
  }
}